import copy
import dataclasses
import uuid
from collections import Counter
from typing import List
from typing import Optional
from typing import Union

import pandas as pd

from evidently.base_metric import InputData
from evidently.model.dashboard import DashboardInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions
from evidently.pipeline.column_mapping import ColumnMapping
from evidently.renderers.base_renderer import TestRenderer
from evidently.suite.base_suite import Display
from evidently.suite.base_suite import Suite
from evidently.suite.base_suite import find_test_renderer
from evidently.test_preset.test_preset import TestPreset
from evidently.tests.base_test import DEFAULT_GROUP
from evidently.tests.base_test import Test
from evidently.tests.base_test import TestResult
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.data_operations import process_columns
from evidently.utils.data_preprocessing import create_data_definition
from evidently.utils.generators import BaseGenerator


class TestSuite(Display):
    _inner_suite: Suite
    _columns_info: DatasetColumns
    _test_presets: List[TestPreset]
    _test_generators: List[BaseGenerator]

    def __init__(
        self,
        tests: Optional[List[Union[Test, TestPreset, BaseGenerator]]],
        options: Optional[list] = None,
    ):
        super().__init__(options)
        self._inner_suite = Suite()
        self._test_presets = []
        self._test_generators = []

        for original_test in tests or []:
            if isinstance(original_test, TestPreset):
                self._test_presets.append(original_test)

            elif isinstance(original_test, BaseGenerator):
                self._test_generators.append(original_test)

            else:
                self._add_test(original_test)

    def _add_test(self, test: Test):
        new_test = copy.copy(test)
        self._inner_suite.add_test(new_test)

    def __bool__(self):
        return all(test_result.is_passed() for _, test_result in self._inner_suite.context.test_results.items())

    def _add_tests_from_generator(self, test_generator: BaseGenerator):
        for test_item in test_generator.generate(columns_info=self._columns_info):
            self._add_test(test_item)

    def run(
        self,
        *,
        reference_data: Optional[pd.DataFrame],
        current_data: pd.DataFrame,
        column_mapping: Optional[ColumnMapping] = None,
    ) -> None:
        if column_mapping is None:
            column_mapping = ColumnMapping()

        self._columns_info = process_columns(current_data, column_mapping)
        data_definition = create_data_definition(reference_data, current_data, column_mapping)
        data = InputData(reference_data, current_data, None, None, column_mapping, data_definition)
        for preset in self._test_presets:
            tests = preset.generate_tests(data, self._columns_info)

            for test in tests:
                if isinstance(test, BaseGenerator):
                    self._add_tests_from_generator(test)

                else:
                    self._add_test(test)

        for test_generator in self._test_generators:
            self._add_tests_from_generator(test_generator)
        self._inner_suite.verify()
        curr_add, ref_add = self._inner_suite.create_additional_features(current_data, reference_data, data_definition)
        data = InputData(reference_data, current_data, ref_add, curr_add, column_mapping, data_definition)

        self._inner_suite.run_calculate(data)
        self._inner_suite.run_checks()

    def as_dict(self) -> dict:
        test_results = []
        counter = Counter(test_result.status for test_result in self._inner_suite.context.test_results.values())

        for test in self._inner_suite.context.test_results:
            renderer = find_test_renderer(type(test), self._inner_suite.context.renderers)
            try:
                test_data = renderer.render_json(test)
                test_results.append(test_data)
            except BaseException as e:
                test_data = TestRenderer.render_json(renderer, test)
                test_data["status"] = TestResult.ERROR
                test_data["description"] = f"Test failed with exception: {e}"
                test_results.append(test_data)

        total_tests = len(self._inner_suite.context.test_results)

        return {
            "tests": test_results,
            "summary": {
                "all_passed": bool(self),
                "total_tests": total_tests,
                "success_tests": counter["SUCCESS"] + counter["WARNING"],
                "failed_tests": counter["FAIL"],
                "by_status": counter,
            },
        }

    def _build_dashboard_info(self):
        test_results = []
        total_tests = len(self._inner_suite.context.test_results)
        by_status = {}
        color_options = self.options_provider.get(ColorOptions)

        for test, test_result in self._inner_suite.context.test_results.items():
            renderer = find_test_renderer(type(test), self._inner_suite.context.renderers)
            renderer.color_options = color_options
            by_status[test_result.status] = by_status.get(test_result.status, 0) + 1
            test_results.append(renderer.render_html(test))

        summary_widget = BaseWidgetInfo(
            title="",
            size=2,
            type="counter",
            params={
                "counters": [{"value": f"{total_tests}", "label": "Tests"}]
                + [
                    {"value": f"{by_status.get(status, 0)}", "label": f"{status.title()}"}
                    for status in [TestResult.SUCCESS, TestResult.WARNING, TestResult.FAIL, TestResult.ERROR]
                ]
            },
        )
        test_suite_widget = BaseWidgetInfo(
            title="",
            type="test_suite",
            size=2,
            params={
                "tests": [
                    dict(
                        title=test_info.name,
                        description=test_info.description,
                        state=test_info.status.lower(),
                        details=dict(
                            parts=[dict(id=item.id, title=item.title, type="widget") for item in test_info.details]
                        ),
                        groups=test_info.groups,
                    )
                    for idx, test_info in enumerate(test_results)
                ],
                "testGroupTypes": DEFAULT_GROUP,
            },
            additionalGraphs=[],
        )
        return (
            "evidently_dashboard_" + str(uuid.uuid4()).replace("-", ""),
            DashboardInfo("Test Suite", widgets=[summary_widget, test_suite_widget]),
            {item.id: dataclasses.asdict(item.info) for idx, info in enumerate(test_results) for item in info.details},
        )



import copy
import dataclasses
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_quality import calculate_correlations
from evidently.features.non_letter_character_percentage_feature import NonLetterCharacterPercentage
from evidently.features.OOV_words_percentage_feature import OOVWordsPercentage
from evidently.features.text_length_feature import TextLength
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import HeatmapData
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import get_heatmaps_widget
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import widget_tabs
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.data_operations import process_columns
from evidently.utils.data_preprocessing import DataDefinition


@dataclasses.dataclass
class CorrelationStats:
    target_prediction_correlation: Optional[float] = None
    abs_max_target_features_correlation: Optional[float] = None
    abs_max_prediction_features_correlation: Optional[float] = None
    abs_max_correlation: Optional[float] = None
    abs_max_features_correlation: Optional[float] = None


@dataclasses.dataclass
class DatasetCorrelation:
    correlation: Dict[str, pd.DataFrame]
    stats: Dict[str, CorrelationStats]
    correlations_calculate: Optional[Dict[str, pd.DataFrame]]


@dataclasses.dataclass
class DatasetCorrelationsMetricResult:
    current: DatasetCorrelation
    reference: Optional[DatasetCorrelation]


class DatasetCorrelationsMetric(Metric[DatasetCorrelationsMetricResult]):
    """Calculate different correlations with target, predictions and features"""

    text_features_gen: Optional[
        Dict[str, Dict[str, Union[TextLength, NonLetterCharacterPercentage, OOVWordsPercentage]]]
    ]

    def __init__(self):
        self.text_features_gen = None

    def required_features(self, data_definition: DataDefinition):
        if len(data_definition.get_columns("text_features")) > 0:
            text_cols = [col.column_name for col in data_definition.get_columns("text_features")]
            text_features_gen = {}
            text_features_gen_result = []
            for col in text_cols:
                col_dict: Dict[str, Union[TextLength, NonLetterCharacterPercentage, OOVWordsPercentage]] = {}
                col_dict[f"{col}: Text Length"] = TextLength(col)
                col_dict[f"{col}: Non Letter Character %"] = NonLetterCharacterPercentage(col)
                col_dict[f"{col}: OOV %"] = OOVWordsPercentage(col)

                text_features_gen_result += [
                    col_dict[f"{col}: Text Length"],
                    col_dict[f"{col}: Non Letter Character %"],
                    col_dict[f"{col}: OOV %"],
                ]
                text_features_gen[col] = col_dict
            self.text_features_gen = text_features_gen

            return text_features_gen_result
        else:
            return []

    def get_parameters(self) -> tuple:
        return ()

    @staticmethod
    def _get_correlations_stats(correlation: pd.DataFrame, columns: DatasetColumns) -> CorrelationStats:
        correlation_matrix = correlation.copy()
        target_name = columns.utility_columns.target
        prediction_name = columns.utility_columns.prediction
        columns_corr = [i for i in correlation_matrix.columns if i not in [target_name, prediction_name]]
        # fill diagonal with 1 values for getting abs max values
        np.fill_diagonal(correlation_matrix.values, 0)

        if (
            isinstance(prediction_name, str)
            and prediction_name in correlation_matrix
            and target_name in correlation_matrix
        ):
            target_prediction_correlation = correlation_matrix.loc[prediction_name, target_name]

            if pd.isnull(target_prediction_correlation):
                target_prediction_correlation = None

        else:
            target_prediction_correlation = None

        if target_name in correlation_matrix:
            abs_max_target_features_correlation = correlation_matrix.loc[target_name, columns_corr].abs().max()

            if pd.isnull(abs_max_target_features_correlation):
                abs_max_target_features_correlation = None

        else:
            abs_max_target_features_correlation = None

        if isinstance(prediction_name, str) and prediction_name in correlation_matrix:
            abs_max_prediction_features_correlation = correlation_matrix.loc[prediction_name, columns_corr].abs().max()

            if pd.isnull(abs_max_prediction_features_correlation):
                abs_max_prediction_features_correlation = None

        else:
            abs_max_prediction_features_correlation = None

        abs_max_correlation = correlation_matrix.abs().max().max()

        if pd.isnull(abs_max_correlation):
            abs_max_correlation = None

        abs_max_features_correlation = correlation_matrix.loc[columns_corr, columns_corr].abs().max().max()

        if pd.isnull(abs_max_features_correlation):
            abs_max_features_correlation = None

        return CorrelationStats(
            target_prediction_correlation=target_prediction_correlation,
            abs_max_target_features_correlation=abs_max_target_features_correlation,
            abs_max_prediction_features_correlation=abs_max_prediction_features_correlation,
            abs_max_correlation=abs_max_correlation,
            abs_max_features_correlation=abs_max_features_correlation,
        )

    def _get_correlations(
        self,
        dataset: pd.DataFrame,
        columns: DatasetColumns,
        add_text_columns: Optional[list],
    ) -> DatasetCorrelation:
        if add_text_columns is not None:
            correlations_calculate = calculate_correlations(dataset, columns, sum(add_text_columns, []))
            correlations = copy.deepcopy(correlations_calculate)
            for name, correlation in correlations_calculate.items():
                if name != "cramer_v":
                    for col_idx in add_text_columns:
                        correlation.loc[col_idx, col_idx] = 0
                    correlations_calculate[name] = correlation
        else:
            correlations_calculate = calculate_correlations(dataset, columns)
            correlations = copy.deepcopy(correlations_calculate)

        stats = {
            name: self._get_correlations_stats(correlation, columns)
            for name, correlation in correlations_calculate.items()
        }

        return DatasetCorrelation(
            correlation=correlations,
            stats=stats,
            correlations_calculate=correlations_calculate,
        )

    def calculate(self, data: InputData) -> DatasetCorrelationsMetricResult:
        columns = process_columns(data.current_data, data.column_mapping)
        curr_df = data.current_data
        ref_df = data.reference_data
        # process text columns
        text_columns = []
        if self.text_features_gen is not None:
            for col in list(self.text_features_gen.keys()):
                curr_text_df = pd.concat(
                    [data.get_current_column(x.feature_name()) for x in list(self.text_features_gen[col].values())],
                    axis=1,
                )
                curr_text_df.columns = list(self.text_features_gen[col].keys())
                text_columns.append(list(curr_text_df.columns))
                curr_df = pd.concat(
                    [curr_df.copy().reset_index(drop=True), curr_text_df.reset_index(drop=True)],
                    axis=1,
                )

                if ref_df is not None:
                    ref_text_df = pd.concat(
                        [
                            data.get_reference_column(x.feature_name())
                            for x in list(self.text_features_gen[col].values())
                        ],
                        axis=1,
                    )
                    ref_text_df.columns = list(self.text_features_gen[col].keys())
                    ref_df = pd.concat(
                        [ref_df.copy().reset_index(drop=True), ref_text_df.reset_index(drop=True)],
                        axis=1,
                    )

        current_correlations = self._get_correlations(dataset=curr_df, columns=columns, add_text_columns=text_columns)

        if ref_df is not None:
            reference_correlation: Optional[DatasetCorrelation] = self._get_correlations(
                dataset=ref_df, columns=columns, add_text_columns=text_columns
            )

        else:
            reference_correlation = None

        return DatasetCorrelationsMetricResult(
            current=current_correlations,
            reference=reference_correlation,
        )


@default_renderer(wrap_type=DatasetCorrelationsMetric)
class DataQualityCorrelationMetricsRenderer(MetricRenderer):
    def render_json(self, obj: DatasetCorrelationsMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        result["current"].pop("correlation", None)
        result["current"].pop("correlations_calculate", None)

        if result["reference"]:
            result["reference"].pop("correlation", None)
            result["reference"].pop("correlations_calculate", None)

        return result

    def _get_heatmaps(self, metric_result: DatasetCorrelationsMetricResult) -> BaseWidgetInfo:
        tabs = []
        curr_corr_result = metric_result.current.correlation
        for correlation_method in curr_corr_result:
            current_correlation = curr_corr_result[correlation_method]

            if metric_result.reference is not None:
                ref_corr_result = metric_result.reference.correlation
                reference_heatmap_data: Optional[HeatmapData] = HeatmapData(
                    name="Reference", matrix=ref_corr_result[correlation_method]
                )

            else:
                reference_heatmap_data = None

            tabs.append(
                TabData(
                    title=correlation_method,
                    widget=get_heatmaps_widget(
                        primary_data=HeatmapData(name="Current", matrix=current_correlation),
                        secondary_data=reference_heatmap_data,
                        color_options=self.color_options,
                    ),
                )
            )

        return widget_tabs(title="", tabs=tabs)

    def render_html(self, obj: DatasetCorrelationsMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        result = [
            header_text(label="Dataset Correlations Metric"),
            self._get_heatmaps(metric_result=metric_result),
        ]
        return result



import dataclasses
from typing import List
from typing import Optional

import numpy as np
import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import header_text
from evidently.utils.data_operations import process_columns
from evidently.utils.visualizations import make_hist_for_num_plot
from evidently.utils.visualizations import plot_distr_subplots


@dataclasses.dataclass
class RegressionErrorDistributionResults:
    current_bins: pd.DataFrame
    reference_bins: Optional[pd.DataFrame]


class RegressionErrorDistribution(Metric[RegressionErrorDistributionResults]):
    def calculate(self, data: InputData) -> RegressionErrorDistributionResults:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction
        curr_df = data.current_data
        ref_df = data.reference_data
        if target_name is None or prediction_name is None:
            raise ValueError("The columns 'target' and 'prediction' columns should be present")
        if not isinstance(prediction_name, str):
            raise ValueError("Expect one column for prediction. List of columns was provided.")
        curr_df = self._make_df_for_plot(curr_df, target_name, prediction_name, None)
        curr_error = curr_df[prediction_name] - curr_df[target_name]
        ref_error = None
        if ref_df is not None:
            ref_df = self._make_df_for_plot(ref_df, target_name, prediction_name, None)
            ref_error = ref_df[prediction_name] - ref_df[target_name]

        result = make_hist_for_num_plot(curr_error, ref_error)
        current_bins = result["current"]
        reference_bins = None
        if "reference" in result.keys():
            reference_bins = result["reference"]

        return RegressionErrorDistributionResults(current_bins=current_bins, reference_bins=reference_bins)

    def _make_df_for_plot(self, df, target_name: str, prediction_name: str, datetime_column_name: Optional[str]):
        result = df.replace([np.inf, -np.inf], np.nan)
        if datetime_column_name is not None:
            result.dropna(axis=0, how="any", inplace=True, subset=[target_name, prediction_name, datetime_column_name])
            return result.sort_values(datetime_column_name)
        result.dropna(axis=0, how="any", inplace=True, subset=[target_name, prediction_name])
        return result.sort_index()


@default_renderer(wrap_type=RegressionErrorDistribution)
class RegressionErrorDistributionRenderer(MetricRenderer):
    def render_json(self, obj: RegressionErrorDistribution) -> dict:
        return {}

    def render_html(self, obj: RegressionErrorDistribution) -> List[BaseWidgetInfo]:
        result = obj.get_result()
        current_bins = result.current_bins
        reference_bins = None
        if result.reference_bins is not None:
            reference_bins = result.reference_bins

        fig = plot_distr_subplots(
            hist_curr=current_bins,
            hist_ref=reference_bins,
            xaxis_name="Error (Predicted - Actual)",
            yaxis_name="Percentage",
            same_color=True,
            color_options=self.color_options,
        )
        return [
            header_text(label="Error Distribution"),
            BaseWidgetInfo(
                title="",
                size=2,
                type="big_graph",
                params={"data": fig["data"], "layout": fig["layout"]},
            ),
        ]



from typing import List
from typing import Optional

from evidently.base_metric import InputData
from evidently.calculations.classification_performance import get_prediction_data
from evidently.metric_preset.metric_preset import MetricPreset
from evidently.metrics import ClassificationClassBalance
from evidently.metrics import ClassificationClassSeparationPlot
from evidently.metrics import ClassificationConfusionMatrix
from evidently.metrics import ClassificationPRCurve
from evidently.metrics import ClassificationProbDistribution
from evidently.metrics import ClassificationPRTable
from evidently.metrics import ClassificationQualityByClass
from evidently.metrics import ClassificationQualityByFeatureTable
from evidently.metrics import ClassificationQualityMetric
from evidently.metrics import ClassificationRocCurve
from evidently.utils.data_operations import DatasetColumns


class ClassificationPreset(MetricPreset):
    """
    Metrics preset for classification performance.

    Contains metrics:
    - ClassificationQualityMetric
    - ClassificationClassBalance
    - ClassificationConfusionMatrix
    - ClassificationQualityByClass
    """

    columns: Optional[List[str]]
    probas_threshold: Optional[float]
    k: Optional[int]

    def __init__(
        self,
        columns: Optional[List[str]] = None,
        probas_threshold: Optional[float] = None,
        k: Optional[int] = None,
    ):
        super().__init__()
        self.columns = columns
        self.probas_threshold = probas_threshold
        self.k = k

    def generate_metrics(self, data: InputData, columns: DatasetColumns):
        result = [
            ClassificationQualityMetric(probas_threshold=self.probas_threshold, k=self.k),
            ClassificationClassBalance(),
            ClassificationConfusionMatrix(probas_threshold=self.probas_threshold, k=self.k),
            ClassificationQualityByClass(probas_threshold=self.probas_threshold, k=self.k),
        ]
        curr_predictions = get_prediction_data(data.current_data, columns, data.column_mapping.pos_label)

        if curr_predictions.prediction_probas is not None:
            result.extend(
                [
                    ClassificationClassSeparationPlot(),
                    ClassificationProbDistribution(),
                    ClassificationRocCurve(),
                    ClassificationPRCurve(),
                    ClassificationPRTable(),
                ]
            )

        result.append(ClassificationQualityByFeatureTable(columns=self.columns))
        return result



import dataclasses
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_quality import ColumnCorrelations
from evidently.calculations.data_quality import calculate_numerical_column_correlations
from evidently.features.non_letter_character_percentage_feature import NonLetterCharacterPercentage
from evidently.features.OOV_words_percentage_feature import OOVWordsPercentage
from evidently.features.text_length_feature import TextLength
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import get_histogram_for_distribution
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import widget_tabs
from evidently.utils.data_operations import process_columns
from evidently.utils.data_preprocessing import ColumnType as ColumnType_data
from evidently.utils.data_preprocessing import DataDefinition


@dataclasses.dataclass
class TextDescriptorsCorrelationMetricResult:
    column_name: str
    current: Dict[str, Dict[str, ColumnCorrelations]]
    reference: Optional[Dict[str, Dict[str, ColumnCorrelations]]] = None


class TextDescriptorsCorrelationMetric(Metric[TextDescriptorsCorrelationMetricResult]):
    """Calculates correlations between each auto-generated text feature for column_name and other dataset columns"""

    column_name: str
    generated_text_features: Dict[str, Union[TextLength, NonLetterCharacterPercentage, OOVWordsPercentage]]

    def __init__(self, column_name: str) -> None:
        self.column_name = column_name
        self.generated_text_features = {}

    def required_features(self, data_definition: DataDefinition):
        column_type = data_definition.get_column(self.column_name).column_type
        if column_type == ColumnType_data.Text:
            self.generated_text_features = {}
            self.generated_text_features["Text Length"] = TextLength(self.column_name)
            self.generated_text_features["Non Letter Character %"] = NonLetterCharacterPercentage(self.column_name)
            self.generated_text_features["OOV %"] = OOVWordsPercentage(self.column_name)
            return list(self.generated_text_features.values())
        return []

    def get_parameters(self) -> tuple:
        return (self.column_name,)

    def calculate(self, data: InputData) -> TextDescriptorsCorrelationMetricResult:
        if self.column_name not in data.current_data:
            raise ValueError(f"Column '{self.column_name}' was not found in current data.")

        if data.reference_data is not None:
            if self.column_name not in data.reference_data:
                raise ValueError(f"Column '{self.column_name}' was not found in reference data.")

        columns = process_columns(data.current_data, data.column_mapping)
        correlation_columns = columns.num_feature_names

        curr_text_df = pd.concat(
            [data.get_current_column(x.feature_name()) for x in list(self.generated_text_features.values())],
            axis=1,
        )
        curr_text_df.columns = list(self.generated_text_features.keys())
        curr_df = pd.concat(
            [data.current_data.copy().reset_index(drop=True), curr_text_df.reset_index(drop=True)],
            axis=1,
        )
        ref_df = None
        if data.reference_data is not None:
            ref_text_df = pd.concat(
                [data.get_reference_column(x.feature_name()) for x in list(self.generated_text_features.values())],
                axis=1,
            )
            ref_text_df.columns = list(self.generated_text_features.keys())
            ref_df = pd.concat(
                [data.reference_data.copy().reset_index(drop=True), ref_text_df.reset_index(drop=True)],
                axis=1,
            )
        curr_result = {}
        ref_result: Optional[dict] = None
        if ref_df is not None:
            ref_result = {}

        for col in list(self.generated_text_features.keys()):
            curr_result[col] = calculate_numerical_column_correlations(col, curr_df, correlation_columns)
            if ref_df is not None and ref_result is not None:
                ref_result[col] = calculate_numerical_column_correlations(col, ref_df, correlation_columns)

        return TextDescriptorsCorrelationMetricResult(
            column_name=self.column_name,
            current=curr_result,
            reference=ref_result,
        )


@default_renderer(wrap_type=TextDescriptorsCorrelationMetric)
class TextDescriptorsCorrelationMetricRenderer(MetricRenderer):
    def render_json(self, obj: TextDescriptorsCorrelationMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        return result

    def _get_plots_correlations(
        self,
        curr_metric_result: Dict,
        ref_metric_result: Optional[Dict],
    ) -> Optional[BaseWidgetInfo]:
        tabs = []

        for correlation_name, current_correlation in curr_metric_result.items():
            reference_correlation_values = None

            if ref_metric_result is not None and correlation_name in ref_metric_result:
                reference_correlation_values = ref_metric_result[correlation_name].values
            # logging.warning(reference_correlation_values)
            if current_correlation.values or reference_correlation_values:
                tabs.append(
                    TabData(
                        title=correlation_name,
                        widget=get_histogram_for_distribution(
                            title="",
                            current_distribution=current_correlation.values,
                            reference_distribution=reference_correlation_values,
                            xaxis_title="Columns",
                            yaxis_title="Correlation",
                            color_options=self.color_options,
                        ),
                    )
                )

        if tabs:
            return widget_tabs(tabs=tabs)

        else:
            return None

    def render_html(self, obj: TextDescriptorsCorrelationMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        result = [header_text(label=f"Correlations for column '{metric_result.column_name}'.")]
        for col in list(metric_result.current.keys()):
            reference = None
            if metric_result.reference is not None:
                reference = metric_result.reference[col]
            correlation_plots = self._get_plots_correlations(metric_result.current[col], reference)
            if correlation_plots:
                result.append(header_text(label=f"{col}"))
                result.append(correlation_plots)
        return result



#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from evidently import ColumnMapping
from evidently.analyzers.prob_classification_performance_analyzer import (
    ProbClassificationPerformanceAnalyzer,
)
from evidently.dashboard.widgets.utils import CutQuantileTransformer
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions
from evidently.options import QualityMetricsOptions


class ProbClassConfusionBasedFeatureDistrTable(Widget):
    def analyzers(self):
        return [ProbClassificationPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = ProbClassificationPerformanceAnalyzer.get_results(analyzers_results)
        quality_metrics_options = self.options_provider.get(QualityMetricsOptions)
        cut_quantile = quality_metrics_options.cut_quantile
        utility_columns = results.columns.utility_columns

        if utility_columns.target is None or utility_columns.prediction is None:
            raise ValueError(
                f"Widget [{self.title}] requires 'target' or 'prediction' columns"
            )

        if current_data is not None:
            ref_array_prediction = reference_data[utility_columns.prediction].to_numpy()
            ref_prediction_ids = np.argmax(ref_array_prediction, axis=-1)
            ref_prediction_labels = [
                utility_columns.prediction[x] for x in ref_prediction_ids
            ]
            reference_data["prediction_labels"] = ref_prediction_labels

            current_array_prediction = current_data[
                utility_columns.prediction
            ].to_numpy()
            current_prediction_ids = np.argmax(current_array_prediction, axis=-1)
            current_prediction_labels = [
                utility_columns.prediction[x] for x in current_prediction_ids
            ]
            current_data["prediction_labels"] = current_prediction_labels

            additional_graphs_data = []
            params_data = []

            for feature_name in results.columns.get_all_features_list(
                cat_before_num=False
            ):
                # add data for table in params
                labels = utility_columns.prediction

                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {"title": "All", "id": "All" + "_" + str(feature_name)}
                            ]
                            + [
                                {
                                    "title": str(label),
                                    "id": feature_name + "_" + str(label),
                                }
                                for label in labels
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                    }
                )

                # create confusion based plots
                reference_data["dataset"] = "Reference"
                current_data["dataset"] = "Current"
                quantile = quality_metrics_options.get_cut_quantile(feature_name)
                if cut_quantile and quantile is not None:
                    side, q = quantile
                    cqt = CutQuantileTransformer(side=side, q=q)
                    cqt.fit(reference_data[feature_name])
                    reference_data_to_plot = cqt.transform_df(
                        reference_data, feature_name
                    )
                    current_data_to_plot = cqt.transform_df(current_data, feature_name)
                else:
                    reference_data_to_plot = reference_data
                    current_data_to_plot = current_data
                merged_data = pd.concat([reference_data_to_plot, current_data_to_plot])

                fig = px.histogram(
                    merged_data,
                    x=feature_name,
                    color=utility_columns.target,
                    facet_col="dataset",
                    histnorm="",
                    barmode="overlay",
                    category_orders={"dataset": ["Reference", "Current"]},
                )

                fig_json = json.loads(fig.to_json())

                # write plot data in table as additional data
                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        "All" + "_" + str(feature_name),
                        {"data": fig_json["data"], "layout": fig_json["layout"]},
                    )
                )

                for label in labels:
                    fig = make_subplots(
                        rows=1, cols=2, subplot_titles=("Reference", "Current")
                    )

                    # REF
                    fig.add_trace(
                        go.Scatter(
                            x=reference_data[
                                reference_data[utility_columns.target] == label
                            ][feature_name],
                            y=reference_data[
                                reference_data[utility_columns.target] == label
                            ][label],
                            mode="markers",
                            name=str(label) + " (ref)",
                            marker=dict(
                                size=6, color=color_options.get_current_data_color()
                            ),
                        ),
                        row=1,
                        col=1,
                    )

                    fig.add_trace(
                        go.Scatter(
                            x=reference_data[
                                reference_data[utility_columns.target] != label
                            ][feature_name],
                            y=reference_data[
                                reference_data[utility_columns.target] != label
                            ][label],
                            mode="markers",
                            name="other (ref)",
                            marker=dict(
                                size=6, color=color_options.get_reference_data_color()
                            ),
                        ),
                        row=1,
                        col=1,
                    )

                    fig.update_layout(
                        xaxis_title=feature_name,
                        yaxis_title="Probability",
                        xaxis=dict(showticklabels=True),
                        yaxis=dict(range=(-0.1, 1.1), showticklabels=True),
                    )

                    # current Prediction
                    fig.add_trace(
                        go.Scatter(
                            x=current_data[
                                current_data[utility_columns.target] == label
                            ][feature_name],
                            y=current_data[
                                current_data[utility_columns.target] == label
                            ][label],
                            mode="markers",
                            name=str(label) + " (curr)",
                            marker=dict(
                                size=6,
                                # set color equal to a variable
                                color=color_options.get_current_data_color(),
                            ),
                        ),
                        row=1,
                        col=2,
                    )

                    fig.add_trace(
                        go.Scatter(
                            x=current_data[
                                current_data[utility_columns.target] != label
                            ][feature_name],
                            y=current_data[
                                current_data[utility_columns.target] != label
                            ][label],
                            mode="markers",
                            name="other (curr)",
                            marker=dict(
                                size=6,
                                # set color equal to a variable
                                color=color_options.get_reference_data_color(),
                            ),
                        ),
                        row=1,
                        col=2,
                    )

                    fig.update_layout(
                        xaxis_title=feature_name,
                        yaxis_title="Probability",
                        xaxis=dict(showticklabels=True),
                        yaxis=dict(range=(-0.1, 1.1), showticklabels=True),
                    )

                    # Update xaxis properties
                    fig.update_xaxes(
                        title_text=feature_name, showgrid=True, row=1, col=1
                    )
                    fig.update_xaxes(
                        title_text=feature_name, showgrid=True, row=1, col=2
                    )

                    # Update yaxis properties
                    fig.update_yaxes(
                        title_text="Probability", showgrid=True, row=1, col=1
                    )
                    fig.update_yaxes(
                        title_text="Probability", showgrid=True, row=1, col=2
                    )

                    fig_json = json.loads(fig.to_json())

                    # write plot data in table as additional data
                    additional_graphs_data.append(
                        AdditionalGraphInfo(
                            feature_name + "_" + str(label),
                            {"data": fig_json["data"], "layout": fig_json["layout"]},
                        )
                    )
        else:
            ref_array_prediction = reference_data[utility_columns.prediction].to_numpy()
            ref_prediction_ids = np.argmax(ref_array_prediction, axis=-1)
            ref_prediction_labels = [
                utility_columns.prediction[x] for x in ref_prediction_ids
            ]
            reference_data["prediction_labels"] = ref_prediction_labels

            additional_graphs_data = []
            params_data = []

            for feature_name in results.columns.get_all_features_list(
                cat_before_num=False
            ):
                # add data for table in params
                labels = utility_columns.prediction

                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {"title": "All", "id": "All" + "_" + str(feature_name)}
                            ]
                            + [
                                {
                                    "title": str(label),
                                    "id": feature_name + "_" + str(label),
                                }
                                for label in labels
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                    }
                )

                # create confusion based plots
                fig = px.histogram(
                    reference_data,
                    x=feature_name,
                    color=utility_columns.target,
                    histnorm="",
                )
                cut_quantile_options = quality_metrics_options.get_cut_quantile(
                    feature_name
                )
                if cut_quantile and cut_quantile_options is not None:
                    side, q = cut_quantile_options
                    cqt = CutQuantileTransformer(side=side, q=q)
                    cqt.fit(reference_data[feature_name])
                    reference_data_to_plot = cqt.transform_df(
                        reference_data, feature_name
                    )
                else:
                    reference_data_to_plot = reference_data

                fig = px.histogram(
                    reference_data_to_plot,
                    x=feature_name,
                    color=utility_columns.target,
                    histnorm="",
                    barmode="overlay",
                )

                fig_json = json.loads(fig.to_json())

                # write plot data in table as additional data
                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        "All" + "_" + str(feature_name),
                        {"data": fig_json["data"], "layout": fig_json["layout"]},
                    )
                )

                for label in labels:
                    fig = go.Figure()

                    fig.add_trace(
                        go.Scatter(
                            x=reference_data[
                                reference_data[utility_columns.target] == label
                            ][feature_name],
                            y=reference_data[
                                reference_data[utility_columns.target] == label
                            ][label],
                            mode="markers",
                            name=str(label),
                            marker=dict(
                                size=6,
                                color=color_options.get_current_data_color(),  # set color equal to a variable
                            ),
                        )
                    )

                    fig.add_trace(
                        go.Scatter(
                            x=reference_data[
                                reference_data[utility_columns.target] != label
                            ][feature_name],
                            y=reference_data[
                                reference_data[utility_columns.target] != label
                            ][label],
                            mode="markers",
                            name="other",
                            marker=dict(
                                size=6, color=color_options.get_reference_data_color()
                            ),
                        )
                    )

                    fig.update_layout(
                        xaxis_title=feature_name,
                        yaxis_title="Probability",
                        xaxis=dict(showticklabels=True),
                        yaxis=dict(range=(-0.1, 1.1), showticklabels=True),
                    )

                    fig_json = json.loads(fig.to_json())

                    # write plot data in table as additional data
                    additional_graphs_data.append(
                        AdditionalGraphInfo(
                            feature_name + "_" + str(label),
                            {"data": fig_json["data"], "layout": fig_json["layout"]},
                        )
                    )

        return BaseWidgetInfo(
            title=self.title,
            type="big_table",
            size=2,
            params={
                "rowsPerPage": min(results.columns.get_features_len(), 10),
                "columns": [{"title": "Feature", "field": "f1"}],
                "data": params_data,
            },
            additionalGraphs=additional_graphs_data,
        )



from typing import Generator

import pandas as pd

from evidently.analyzers.data_quality_analyzer import DataQualityAnalyzer
from evidently.calculations.data_quality import DataQualityStats
from evidently.model_monitoring.monitoring import MetricsType
from evidently.model_monitoring.monitoring import ModelMonitor
from evidently.model_monitoring.monitoring import ModelMonitoringMetric


class DataQualityMonitorMetrics:
    _tag = "data_quality"
    quality_stat = ModelMonitoringMetric(
        f"{_tag}:quality_stat", ["dataset", "feature", "feature_type", "metric"]
    )


class DataQualityMonitor(ModelMonitor):
    def monitor_id(self) -> str:
        return "data_quality"

    def analyzers(self):
        return [DataQualityAnalyzer]

    @staticmethod
    def _yield_metrics(
        data_stats: DataQualityStats,
        dataset: str,
    ) -> Generator[MetricsType, None, None]:
        for feature_name, feature_stats in data_stats.get_all_features().items():
            feature_stats_dict = feature_stats.as_dict()
            feature_type = feature_stats_dict.pop("feature_type")
            for stat_name, stat_value in feature_stats_dict.items():
                if stat_value is not None:
                    if pd.isnull(stat_value):
                        stat_value = None

                    yield DataQualityMonitorMetrics.quality_stat.create(
                        stat_value,
                        {
                            "dataset": dataset,
                            "feature": feature_name,
                            "feature_type": feature_type,
                            "metric": stat_name,
                        },
                    )

    def metrics(self, analyzer_results) -> Generator[MetricsType, None, None]:
        results = DataQualityAnalyzer.get_results(analyzer_results)

        if results.reference_features_stats is not None:
            for metric in self._yield_metrics(
                results.reference_features_stats, "reference"
            ):
                yield metric

        if results.current_features_stats is not None:
            for metric in self._yield_metrics(
                results.current_features_stats, "current"
            ):
                yield metric



"""Available statistical tests.
For detailed information about statistical tests see module documentation.
"""
from .anderson_darling_stattest import anderson_darling_test
from .chisquare_stattest import chi_stat_test
from .cramer_von_mises_stattest import cramer_von_mises
from .energy_distance import energy_dist_test
from .epps_singleton_stattest import epps_singleton_test
from .fisher_exact_stattest import fisher_exact_test
from .g_stattest import g_test
from .hellinger_distance import hellinger_stat_test
from .jensenshannon import jensenshannon_stat_test
from .kl_div import kl_div_stat_test
from .ks_stattest import ks_stat_test
from .mann_whitney_urank_stattest import mann_whitney_u_stat_test
from .psi import psi_stat_test
from .registry import PossibleStatTestType
from .registry import StatTest
from .registry import StatTestFuncType
from .registry import get_stattest
from .registry import register_stattest
from .t_test import t_test
from .text_content_drift import text_content_drift_stat_test
from .tvd_stattest import tvd_test
from .wasserstein_distance_norm import wasserstein_stat_test
from .z_stattest import z_stat_test



import dataclasses
from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.analyzers.base_analyzer import Analyzer
from evidently.analyzers.base_analyzer import BaseAnalyzerResult
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class ProbDistributionAnalyzerResults(BaseAnalyzerResult):
    pass


# TODO: Move logic from `evidently.dashboard.widgets.prob_class_pred_distr_widget.py` to here
class ProbDistributionAnalyzer(Analyzer):
    @staticmethod
    def get_results(analyzer_results) -> ProbDistributionAnalyzerResults:
        return analyzer_results[ProbDistributionAnalyzer]

    def calculate(
        self, reference_data: pd.DataFrame, current_data: Optional[pd.DataFrame], column_mapping: ColumnMapping
    ) -> ProbDistributionAnalyzerResults:
        columns = process_columns(reference_data, column_mapping)

        return ProbDistributionAnalyzerResults(columns=columns)



import collections
import dataclasses
import re
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Pattern

import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_quality import get_rows_count
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import table_data
from evidently.renderers.html_widgets import widget_tabs


@dataclass
class DataIntegrityValueByRegexpStat:
    """Statistics about matched by a regular expression values in a column for one dataset"""

    # count of matched values in the column, without NaNs
    number_of_matched: int
    # count of not matched values in the column, without NaNs
    number_of_not_matched: int
    # count of rows in the column, including matched, not matched and NaNs
    number_of_rows: int
    # map with matched values (keys) and count of the values (value)
    table_of_matched: Dict[str, int]
    # map with not matched values (keys) and count of the values (values)
    table_of_not_matched: Dict[str, int]


@dataclass
class DataIntegrityValueByRegexpMetricResult:
    # name of the column that we check by the regular expression
    column_name: str
    # the regular expression as a string
    reg_exp: str
    top: int
    # match statistic for current dataset
    current: DataIntegrityValueByRegexpStat
    # match statistic for reference dataset, equals None if the reference is not present
    reference: Optional[DataIntegrityValueByRegexpStat] = None


class ColumnRegExpMetric(Metric[DataIntegrityValueByRegexpMetricResult]):
    """Count number of values in a column matched or not by a regular expression (regexp)"""

    # name of the column that we check
    column_name: str
    # the regular expression
    reg_exp: str
    top: int
    # compiled regular expression for speed optimization
    _reg_exp_compiled: Pattern

    def __init__(self, column_name: str, reg_exp: str, top: int = 10):
        self.top = top
        self.reg_exp = reg_exp
        self.column_name = column_name
        self._reg_exp_compiled = re.compile(reg_exp)

    def _calculate_stats_by_regexp(self, column: pd.Series) -> DataIntegrityValueByRegexpStat:
        number_of_matched = 0
        number_of_na = 0
        number_of_not_matched = 0
        table_of_matched: Dict[str, int] = collections.defaultdict(int)
        table_of_not_matched: Dict[str, int] = collections.defaultdict(int)

        for item in column:
            if pd.isna(item):
                number_of_na += 1
                continue

            item = str(item)

            if bool(self._reg_exp_compiled.match(str(item))):
                number_of_matched += 1
                table_of_matched[item] += 1

            else:
                number_of_not_matched += 1
                table_of_not_matched[item] += 1
        matched = sorted(table_of_matched.items(), key=lambda x: x[1], reverse=True)
        table_of_matched = {k: v for k, v in matched[: self.top]}

        not_matched = sorted(table_of_not_matched.items(), key=lambda x: x[1], reverse=True)
        table_of_not_matched = {k: v for k, v in not_matched[: self.top]}

        return DataIntegrityValueByRegexpStat(
            number_of_matched=number_of_matched,
            number_of_not_matched=number_of_not_matched,
            number_of_rows=get_rows_count(column),
            table_of_matched=table_of_matched,
            table_of_not_matched=table_of_not_matched,
        )

    def calculate(self, data: InputData) -> DataIntegrityValueByRegexpMetricResult:
        if self.top < 1:
            raise ValueError("Parameter top must be >= 1")

        if not self.reg_exp:
            raise ValueError("Parameter reg_exp must be not empty for ColumnRegExpMetric")

        if self.column_name not in data.current_data:
            raise ValueError(f"Column {self.column_name} not found in current dataset.")

        current = self._calculate_stats_by_regexp(data.current_data[self.column_name])
        reference = None

        if data.reference_data is not None:
            if self.column_name not in data.reference_data:
                raise ValueError(f"Column {self.column_name} was not found in reference dataset.")

            reference = self._calculate_stats_by_regexp(data.reference_data[self.column_name])

        return DataIntegrityValueByRegexpMetricResult(
            column_name=self.column_name, reg_exp=self.reg_exp, top=self.top, current=current, reference=reference
        )


@default_renderer(wrap_type=ColumnRegExpMetric)
class ColumnRegExpMetricRenderer(MetricRenderer):
    def render_json(self, obj: ColumnRegExpMetric) -> dict:
        return dataclasses.asdict(obj.get_result())

    @staticmethod
    def _get_counters(dataset_name: str, metrics: DataIntegrityValueByRegexpStat) -> BaseWidgetInfo:
        percents = round(metrics.number_of_not_matched * 100 / metrics.number_of_rows, 3)
        counters = [
            CounterData(label="Number of Values", value=f"{metrics.number_of_rows}"),
            CounterData(label="Mismatched", value=f"{metrics.number_of_not_matched} ({percents}%)"),
        ]
        return counter(
            counters=counters,
            title=f"{dataset_name.capitalize()} dataset",
        )

    @staticmethod
    def _get_table_stat(dataset_name: str, top: int, metrics: DataIntegrityValueByRegexpStat) -> BaseWidgetInfo:
        return table_data(
            title=f"{dataset_name.capitalize()} Dataset: top {top} mismatched values",
            column_names=["Value", "Count"],
            data=metrics.table_of_not_matched.items(),
        )

    def render_html(self, obj: ColumnRegExpMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        column_name = metric_result.column_name

        result = [
            header_text(label=f"RegExp Match for column '{column_name}'."),
            self._get_counters("current", metric_result.current),
        ]

        if metric_result.reference is not None:
            result.append(self._get_counters("reference", metric_result.reference))

        current_table = self._get_table_stat("current", metric_result.top, metric_result.current)

        if metric_result.reference is not None:
            tables_tabs = [
                TabData(title="Current dataset", widget=current_table),
                TabData(
                    title="Reference dataset",
                    widget=self._get_table_stat("reference", metric_result.top, metric_result.reference),
                ),
            ]
            tables = widget_tabs(tabs=tables_tabs)

        else:
            tables = current_table

        result.append(tables)
        return result



import abc
from abc import ABC
from typing import Any
from typing import List
from typing import Optional
from typing import Union

from evidently.calculations.classification_performance import DatasetClassificationQuality
from evidently.metrics.classification_performance.classification_dummy_metric import ClassificationDummyMetric
from evidently.metrics.classification_performance.classification_quality_metric import ClassificationConfusionMatrix
from evidently.metrics.classification_performance.classification_quality_metric import ClassificationQualityMetric
from evidently.metrics.classification_performance.quality_by_class_metric import ClassificationQualityByClass
from evidently.metrics.classification_performance.roc_curve_metric import ClassificationRocCurve
from evidently.renderers.base_renderer import TestHtmlInfo
from evidently.renderers.base_renderer import TestRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import get_roc_auc_tab_data
from evidently.renderers.html_widgets import plotly_figure
from evidently.renderers.html_widgets import widget_tabs
from evidently.tests.base_test import BaseCheckValueTest
from evidently.tests.base_test import GroupData
from evidently.tests.base_test import GroupingTypes
from evidently.tests.base_test import TestValueCondition
from evidently.tests.utils import approx
from evidently.tests.utils import plot_boxes
from evidently.tests.utils import plot_conf_mtrx
from evidently.tests.utils import plot_rates
from evidently.utils.types import Numeric

CLASSIFICATION_GROUP = GroupData("classification", "Classification", "")
GroupingTypes.TestGroup.add_value(CLASSIFICATION_GROUP)


class SimpleClassificationTest(BaseCheckValueTest):
    group = CLASSIFICATION_GROUP.id
    name: str
    metric: ClassificationQualityMetric
    dummy_metric: ClassificationDummyMetric

    def __init__(
        self,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.metric = ClassificationQualityMetric()
        self.dummy_metric = ClassificationDummyMetric()

    def calculate_value_for_test(self) -> Optional[Any]:
        return self.get_value(self.metric.get_result().current)

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        ref_metrics = self.metric.get_result().reference

        if ref_metrics is not None:
            return TestValueCondition(eq=approx(self.get_value(ref_metrics), relative=0.2))

        if self.get_value(self.dummy_metric.get_result().dummy) is None:
            raise ValueError("Neither required test parameters nor reference data has been provided.")

        return TestValueCondition(gt=self.get_value(self.dummy_metric.get_result().dummy))

    @abc.abstractmethod
    def get_value(self, result: DatasetClassificationQuality):
        raise NotImplementedError()


class SimpleClassificationTestTopK(SimpleClassificationTest, ABC):
    metric: ClassificationQualityMetric
    dummy_metric: ClassificationDummyMetric
    conf_matrix: ClassificationConfusionMatrix
    probas_threshold: Optional[float]
    k: Optional[Union[float, int]]

    def __init__(
        self,
        probas_threshold: Optional[float] = None,
        k: Optional[Union[float, int]] = None,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        super().__init__(
            eq=eq,
            gt=gt,
            gte=gte,
            is_in=is_in,
            lt=lt,
            lte=lte,
            not_eq=not_eq,
            not_in=not_in,
        )
        if k is not None and probas_threshold is not None:
            raise ValueError("Only one of 'probas_threshold' or 'k' should be given")
        self.k = k
        self.probas_threshold = probas_threshold
        self.dummy_metric = ClassificationDummyMetric(probas_threshold=self.probas_threshold, k=self.k)
        self.metric = ClassificationQualityMetric(probas_threshold=self.probas_threshold, k=self.k)
        self.conf_matrix = ClassificationConfusionMatrix(probas_threshold=self.probas_threshold, k=self.k)

    def calculate_value_for_test(self) -> Optional[Any]:
        return self.get_value(self.metric.get_result().current)

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        result = self.metric.get_result()
        ref_metrics = result.reference

        if ref_metrics is not None:
            return TestValueCondition(eq=approx(self.get_value(ref_metrics), relative=0.2))

        dummy_result = self.dummy_metric.get_result().dummy

        if self.get_value(dummy_result) is None:
            raise ValueError("Neither required test parameters nor reference data has been provided.")

        return TestValueCondition(gt=self.get_value(dummy_result))


class TestAccuracyScore(SimpleClassificationTestTopK):
    name = "Accuracy Score"

    def get_value(self, result: DatasetClassificationQuality):
        return result.accuracy

    def get_description(self, value: Numeric) -> str:
        return f"The Accuracy Score is {value:.3g}. The test threshold is {self.get_condition()}"


@default_renderer(wrap_type=TestAccuracyScore)
class TestAccuracyScoreRenderer(TestRenderer):
    def render_json(self, obj: TestAccuracyScore) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["accuracy"] = obj.value
        return base

    def render_html(self, obj: TestAccuracyScore) -> TestHtmlInfo:
        info = super().render_html(obj)
        curr_matrix = obj.conf_matrix.get_result().current_matrix
        ref_matrix = obj.conf_matrix.get_result().reference_matrix
        fig = plot_conf_mtrx(curr_matrix, ref_matrix)
        info.with_details("Accuracy Score", plotly_figure(figure=fig, title=""))
        return info


class TestPrecisionScore(SimpleClassificationTestTopK):
    name = "Precision Score"

    def get_value(self, result: DatasetClassificationQuality):
        return result.precision

    def get_description(self, value: Numeric) -> str:
        return f"The Precision Score is {value:.3g}. The test threshold is {self.get_condition()}"


@default_renderer(wrap_type=TestPrecisionScore)
class TestPrecisionScoreRenderer(TestRenderer):
    def render_json(self, obj: TestPrecisionScore) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["precision"] = obj.value
        return base

    def render_html(self, obj: TestPrecisionScore) -> TestHtmlInfo:
        info = super().render_html(obj)
        curr_matrix = obj.conf_matrix.get_result().current_matrix
        ref_matrix = obj.conf_matrix.get_result().reference_matrix
        fig = plot_conf_mtrx(curr_matrix, ref_matrix)
        info.with_details("Precision Score", plotly_figure(figure=fig, title=""))
        return info


class TestF1Score(SimpleClassificationTestTopK):
    name = "F1 Score"

    def get_value(self, result: DatasetClassificationQuality):
        return result.f1

    def get_description(self, value: Numeric) -> str:
        return f"The F1 Score is {value:.3g}. The test threshold is {self.get_condition()}"


@default_renderer(wrap_type=TestF1Score)
class TestF1ScoreRenderer(TestRenderer):
    def render_json(self, obj: TestF1Score) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["f1"] = obj.value
        return base

    def render_html(self, obj: TestF1Score) -> TestHtmlInfo:
        info = super().render_html(obj)
        curr_matrix = obj.conf_matrix.get_result().current_matrix
        ref_matrix = obj.conf_matrix.get_result().reference_matrix
        fig = plot_conf_mtrx(curr_matrix, ref_matrix)
        info.with_details("F1 Score", plotly_figure(title="", figure=fig))
        return info


class TestRecallScore(SimpleClassificationTestTopK):
    name = "Recall Score"

    def get_value(self, result: DatasetClassificationQuality):
        return result.recall

    def get_description(self, value: Numeric) -> str:
        return f"The Recall Score is {value:.3g}. The test threshold is {self.get_condition()}"


@default_renderer(wrap_type=TestRecallScore)
class TestRecallScoreRenderer(TestRenderer):
    def render_json(self, obj: TestRecallScore) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["recall"] = obj.value
        return base

    def render_html(self, obj: TestRecallScore) -> TestHtmlInfo:
        info = super().render_html(obj)
        curr_matrix = obj.conf_matrix.get_result().current_matrix
        ref_matrix = obj.conf_matrix.get_result().reference_matrix
        fig = plot_conf_mtrx(curr_matrix, ref_matrix)
        info.with_details("Recall Score", plotly_figure(title="", figure=fig))
        return info


class TestRocAuc(SimpleClassificationTest):
    name = "ROC AUC Score"
    roc_curve: ClassificationRocCurve

    def __init__(
        self,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.roc_curve = ClassificationRocCurve()

    def get_value(self, result: DatasetClassificationQuality):
        return result.roc_auc

    def get_description(self, value: Numeric) -> str:
        if value is None:
            return "Not enough data to calculate ROC AUC. Consider providing probabilities instead of labels."

        else:
            return f"The ROC AUC Score is {value:.3g}. The test threshold is {self.get_condition()}"


@default_renderer(wrap_type=TestRocAuc)
class TestRocAucRenderer(TestRenderer):
    def render_json(self, obj: TestRocAuc) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["roc_auc"] = obj.value
        return base

    def render_html(self, obj: TestRocAuc) -> TestHtmlInfo:
        info = super().render_html(obj)
        curr_roc_curve = obj.roc_curve.get_result().current_roc_curve
        ref_roc_curve = obj.roc_curve.get_result().reference_roc_curve

        if curr_roc_curve is None:
            return info

        tab_data = get_roc_auc_tab_data(curr_roc_curve, ref_roc_curve, color_options=self.color_options)

        if len(tab_data) == 1:
            return info.with_details("ROC Curve", tab_data[0][1])

        tabs = [TabData(name, widget) for name, widget in tab_data]
        return info.with_details("", widget_tabs(title="", tabs=tabs))


class TestLogLoss(SimpleClassificationTest):
    name = "Logarithmic Loss"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        ref_metrics = self.metric.get_result().reference

        if ref_metrics is not None:
            return TestValueCondition(eq=approx(self.get_value(ref_metrics), relative=0.2))

        if self.get_value(self.dummy_metric.get_result().dummy) is None:
            raise ValueError("Neither required test parameters nor reference data has been provided.")

        return TestValueCondition(lt=self.get_value(self.dummy_metric.get_result().dummy))

    def get_value(self, result: DatasetClassificationQuality):
        return result.log_loss

    def get_description(self, value: Numeric) -> str:
        if value is None:
            return "Not enough data to calculate Logarithmic Loss. Consider providing probabilities instead of labels."

        else:
            return f"The Logarithmic Loss is {value:.3g}. The test threshold is {self.get_condition()}"


@default_renderer(wrap_type=TestLogLoss)
class TestLogLossRenderer(TestRenderer):
    def render_json(self, obj: TestLogLoss) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["log_loss"] = obj.value
        return base

    def render_html(self, obj: TestLogLoss) -> TestHtmlInfo:
        info = super().render_html(obj)
        result = obj.metric.get_result()

        curr_metrics = result.current.plot_data
        ref_metrics = None if result.reference is None else result.reference.plot_data

        if curr_metrics is not None:
            fig = plot_boxes(curr_for_plots=curr_metrics, ref_for_plots=ref_metrics, color_options=self.color_options)
            info.with_details("Logarithmic Loss", plotly_figure(title="", figure=fig))

        return info


class TestTPR(SimpleClassificationTestTopK):
    name = "True Positive Rate"

    def get_value(self, result: DatasetClassificationQuality):
        return result.tpr

    def get_description(self, value: Numeric) -> str:
        if value is None:
            return "This test is applicable only for binary classification"

        return f"The True Positive Rate is {value:.3g}. The test threshold is {self.get_condition()}"


@default_renderer(wrap_type=TestTPR)
class TestTPRRenderer(TestRenderer):
    def render_json(self, obj: TestTPR) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["tpr"] = obj.value
        return base

    def render_html(self, obj: TestF1Score) -> TestHtmlInfo:
        info = super().render_html(obj)
        curr_metrics = obj.metric.get_result().current
        ref_metrics = obj.metric.get_result().reference
        curr_rate_plots_data = curr_metrics.rate_plots_data
        ref_rate_plots_data = None

        if ref_metrics is not None:
            ref_rate_plots_data = ref_metrics.rate_plots_data

        if curr_rate_plots_data is not None:
            fig = plot_rates(
                curr_rate_plots_data=curr_rate_plots_data,
                ref_rate_plots_data=ref_rate_plots_data,
                color_options=self.color_options,
            )
            info.with_details("TPR", plotly_figure(title="", figure=fig))

        return info


class TestTNR(SimpleClassificationTestTopK):
    name = "True Negative Rate"

    def get_value(self, result: DatasetClassificationQuality):
        return result.tnr

    def get_description(self, value: Numeric) -> str:
        if value is None:
            return "This test is applicable only for binary classification"

        return f"The True Negative Rate is {value:.3g}. The test threshold is {self.get_condition()}"


@default_renderer(wrap_type=TestTNR)
class TestTNRRenderer(TestRenderer):
    def render_json(self, obj: TestTNR) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["tnr"] = obj.value
        return base

    def render_html(self, obj: TestF1Score) -> TestHtmlInfo:
        info = super().render_html(obj)
        curr_metrics = obj.metric.get_result().current
        ref_metrics = obj.metric.get_result().reference
        curr_rate_plots_data = curr_metrics.rate_plots_data
        ref_rate_plots_data = None

        if ref_metrics is not None:
            ref_rate_plots_data = ref_metrics.rate_plots_data

        if curr_rate_plots_data is not None:
            fig = plot_rates(
                curr_rate_plots_data=curr_rate_plots_data,
                ref_rate_plots_data=ref_rate_plots_data,
                color_options=self.color_options,
            )
            info.with_details("TNR", plotly_figure(title="", figure=fig))

        return info


class TestFPR(SimpleClassificationTestTopK):
    name = "False Positive Rate"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        result = self.metric.get_result()
        ref_metrics = result.reference
        dummy_metrics = self.dummy_metric.get_result().dummy

        if ref_metrics is not None:
            return TestValueCondition(eq=approx(self.get_value(ref_metrics), relative=0.2))

        if self.get_value(dummy_metrics) is None:
            raise ValueError("Neither required test parameters nor reference data has been provided.")

        return TestValueCondition(lt=self.get_value(dummy_metrics))

    def get_value(self, result: DatasetClassificationQuality):
        return result.fpr

    def get_description(self, value: Numeric) -> str:
        if value is None:
            return "This test is applicable only for binary classification"

        return f"The False Positive Rate is {value:.3g}. The test threshold is {self.get_condition()}"


@default_renderer(wrap_type=TestFPR)
class TestFPRRenderer(TestRenderer):
    def render_json(self, obj: TestFPR) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["fpr"] = obj.value
        return base

    def render_html(self, obj: TestF1Score) -> TestHtmlInfo:
        info = super().render_html(obj)
        curr_metrics = obj.metric.get_result().current
        ref_metrics = obj.metric.get_result().reference
        curr_rate_plots_data = curr_metrics.rate_plots_data
        ref_rate_plots_data = None

        if ref_metrics is not None:
            ref_rate_plots_data = ref_metrics.rate_plots_data

        if curr_rate_plots_data is not None:
            fig = plot_rates(
                curr_rate_plots_data=curr_rate_plots_data,
                ref_rate_plots_data=ref_rate_plots_data,
                color_options=self.color_options,
            )
            info.with_details("FPR", plotly_figure(title="", figure=fig))

        return info


class TestFNR(SimpleClassificationTestTopK):
    name = "False Negative Rate"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        result = self.metric.get_result()
        ref_metrics = result.reference
        dummy_metrics = self.dummy_metric.get_result().dummy

        if ref_metrics is not None:
            return TestValueCondition(eq=approx(self.get_value(ref_metrics), relative=0.2))

        if self.get_value(dummy_metrics) is None:
            raise ValueError("Neither required test parameters nor reference data has been provided.")

        return TestValueCondition(lt=self.get_value(dummy_metrics))

    def get_value(self, result: DatasetClassificationQuality):
        return result.fnr

    def get_description(self, value: Numeric) -> str:
        if value is None:
            return "This test is applicable only for binary classification"

        return f"The False Negative Rate is {value:.3g}. The test threshold is {self.get_condition()}"


@default_renderer(wrap_type=TestFNR)
class TestFNRRenderer(TestRenderer):
    def render_json(self, obj: TestFNR) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["fnr"] = obj.value
        return base

    def render_html(self, obj: TestF1Score) -> TestHtmlInfo:
        info = super().render_html(obj)
        curr_metrics = obj.metric.get_result().current
        ref_metrics = obj.metric.get_result().reference
        curr_rate_plots_data = curr_metrics.rate_plots_data
        ref_rate_plots_data = None

        if ref_metrics is not None:
            ref_rate_plots_data = ref_metrics.rate_plots_data

        if curr_rate_plots_data is not None:
            fig = plot_rates(
                curr_rate_plots_data=curr_rate_plots_data,
                ref_rate_plots_data=ref_rate_plots_data,
                color_options=self.color_options,
            )
            info.with_details("FNR", plotly_figure(title="", figure=fig))

        return info


class ByClassClassificationTest(BaseCheckValueTest, ABC):
    group = CLASSIFICATION_GROUP.id
    metric: ClassificationQualityMetric
    by_class_metric: ClassificationQualityByClass
    dummy_metric: ClassificationDummyMetric
    conf_matrix: ClassificationConfusionMatrix

    def __init__(
        self,
        label: str,
        probas_threshold: Optional[float] = None,
        k: Optional[Union[float, int]] = None,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)

        if k is not None and probas_threshold is not None:
            raise ValueError("Only one of 'probas_threshold' or 'k' should be given")

        self.label = label
        self.probas_threshold = probas_threshold
        self.k = k
        self.metric = ClassificationQualityMetric(probas_threshold=self.probas_threshold, k=self.k)
        self.dummy_metric = ClassificationDummyMetric(probas_threshold=self.probas_threshold, k=self.k)
        self.by_class_metric = ClassificationQualityByClass(probas_threshold=self.probas_threshold, k=self.k)
        self.conf_matrix = ClassificationConfusionMatrix(probas_threshold=self.probas_threshold, k=self.k)

    def calculate_value_for_test(self) -> Optional[Any]:
        return self.get_value(self.by_class_metric.get_result().current_metrics[self.label])

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        result = self.by_class_metric.get_result()
        ref_metrics = result.reference_metrics

        if ref_metrics is not None:
            return TestValueCondition(eq=approx(self.get_value(ref_metrics[self.label]), relative=0.2))

        dummy_result = self.dummy_metric.get_result().metrics_matrix[self.label]

        if self.get_value(dummy_result) is None:
            raise ValueError("Neither required test parameters nor reference data has been provided.")

        return TestValueCondition(gt=self.get_value(dummy_result))

    @abc.abstractmethod
    def get_value(self, result: dict):
        raise NotImplementedError()


class TestPrecisionByClass(ByClassClassificationTest):
    name: str = "Precision Score by Class"

    def get_value(self, result: dict):
        return result["precision"]

    def get_description(self, value: Numeric) -> str:
        return (
            f"The precision score of the label **{self.label}** is {value:.3g}. "
            f"The test threshold is {self.get_condition()}"
        )


@default_renderer(wrap_type=TestPrecisionByClass)
class TestPrecisionByClassRenderer(TestRenderer):
    def render_json(self, obj: TestPrecisionByClass) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["precision"] = obj.value
        base["parameters"]["label"] = obj.label
        return base

    def render_html(self, obj: TestPrecisionByClass) -> TestHtmlInfo:
        info = super().render_html(obj)
        curr_matrix = obj.conf_matrix.get_result().current_matrix
        ref_matrix = obj.conf_matrix.get_result().reference_matrix
        fig = plot_conf_mtrx(curr_matrix, ref_matrix)
        info.with_details("Precision by Class", plotly_figure(title="", figure=fig))
        return info


class TestRecallByClass(ByClassClassificationTest):
    name: str = "Recall Score by Class"

    def get_value(self, result: dict):
        return result["recall"]

    def get_description(self, value: Numeric) -> str:
        return (
            f"The recall score of the label **{self.label}** is {value:.3g}. "
            f"The test threshold is {self.get_condition()}"
        )


@default_renderer(wrap_type=TestRecallByClass)
class TestRecallByClassRenderer(TestRenderer):
    def render_json(self, obj: TestRecallByClass) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["recall"] = obj.value
        base["parameters"]["label"] = obj.label
        return base

    def render_html(self, obj: TestRecallByClass) -> TestHtmlInfo:
        info = super().render_html(obj)
        curr_matrix = obj.conf_matrix.get_result().current_matrix
        ref_matrix = obj.conf_matrix.get_result().reference_matrix
        fig = plot_conf_mtrx(curr_matrix, ref_matrix)
        info.with_details("Recall by Class", plotly_figure(title="", figure=fig))
        return info


class TestF1ByClass(ByClassClassificationTest):
    name: str = "F1 Score by Class"

    def get_value(self, result: dict):
        return result["f1-score"]

    def get_description(self, value: Numeric) -> str:
        return (
            f"The F1 score of the label **{self.label}** is {value:.3g}. The test threshold is {self.get_condition()}"
        )


@default_renderer(wrap_type=TestF1ByClass)
class TestF1ByClassRenderer(TestRenderer):
    def render_json(self, obj: TestF1ByClass) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["f1"] = obj.value
        base["parameters"]["label"] = obj.label
        return base

    def render_html(self, obj: TestF1ByClass) -> TestHtmlInfo:
        info = super().render_html(obj)
        curr_matrix = obj.conf_matrix.get_result().current_matrix
        ref_matrix = obj.conf_matrix.get_result().reference_matrix
        fig = plot_conf_mtrx(curr_matrix, ref_matrix)
        info.with_details("F1 by Class", plotly_figure(title="", figure=fig))
        return info



from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import pandas as pd
import plotly.graph_objs as go
from pandas.api.types import is_numeric_dtype
from plotly.subplots import make_subplots

from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions
from evidently.renderers.base_renderer import DetailsInfo
from evidently.renderers.html_widgets import plotly_figure
from evidently.renderers.html_widgets import table_data
from evidently.utils.types import ApproxValue


def plot_check(fig, condition, color_options: ColorOptions):
    lines = []
    left_line = pd.Series([condition.gt, condition.gte]).max()

    if not pd.isnull(left_line):
        left_line_name = ["gt", "gte"][pd.Series([condition.gt, condition.gte]).argmax()]
        lines.append((left_line, left_line_name))

    right_line = pd.Series([condition.lt, condition.lte]).min()

    if not pd.isnull(right_line):
        right_line_name = ["lt", "lte"][pd.Series([condition.lt, condition.lte]).argmin()]
        lines.append((right_line, right_line_name))

    if condition.eq is not None and not isinstance(condition.eq, ApproxValue):
        lines.append((condition.eq, "eq"))

    if condition.eq is not None and isinstance(condition.eq, ApproxValue):
        lines.append((condition.eq.value, "approx"))

    if condition.not_eq is not None:
        lines.append((condition.not_eq, "not_eq"))

    fig = go.Figure(fig)
    max_y = np.max([np.max(x["y"]) for x in fig.data])

    if len(lines) > 0:
        for line, name in lines:
            fig.add_trace(
                go.Scatter(
                    x=(line, line),
                    y=(0, max_y),
                    mode="lines",
                    line=dict(color=color_options.secondary_color, width=3, dash="dash"),
                    name=name,
                )
            )

    if left_line and right_line:
        fig.add_vrect(x0=left_line, x1=right_line, fillcolor="green", opacity=0.25, line_width=0)

    if condition.eq is not None and isinstance(condition.eq, ApproxValue):
        left_border = 0.0
        right_border = 0.0

        if condition.eq._relative > 1e-6:
            left_border = condition.eq.value - condition.eq.value * condition.eq._relative
            right_border = condition.eq.value + condition.eq.value * condition.eq._relative
            fig.add_vrect(x0=left_border, x1=right_border, fillcolor="green", opacity=0.25, line_width=0)

        elif condition.eq._absolute > 1e-12:
            left_border = condition.eq.value - condition.eq._absolute
            right_border = condition.eq.value + condition.eq._absolute
            fig.add_vrect(x0=left_border, x1=right_border, fillcolor="green", opacity=0.25, line_width=0)

        fig.add_vrect(x0=left_border, x1=right_border, fillcolor="green", opacity=0.25, line_width=0)

    fig.update_layout(showlegend=True)
    return fig


def plot_metric_value(fig, metric_val: float, metric_name: str):
    fig = go.Figure(fig)
    max_y = np.max([np.max(x["y"]) for x in fig.data])
    min_y = np.min([np.min(x["y"]) for x in fig.data])
    fig.add_trace(
        go.Scatter(
            x=(metric_val, metric_val),
            y=(min_y, max_y),
            mode="lines",
            line=dict(color="green", width=3),
            name=metric_name,
        )
    )
    fig.update_layout(showlegend=True)
    return fig


def regression_perf_plot(
    *,
    val_for_plot: Dict[str, pd.Series],
    hist_for_plot: Dict[str, pd.Series],
    name: str,
    curr_metric: float,
    ref_metric: float = None,
    is_ref_data: bool = False,
    color_options: ColorOptions,
):
    current_color = color_options.get_current_data_color()
    reference_color = color_options.get_reference_data_color()
    fig = make_subplots(rows=2, cols=1, shared_xaxes=True)
    sorted_index = val_for_plot["current"].sort_index()
    x = [str(idx) for idx in sorted_index.index]
    y = list(sorted_index)
    trace = go.Scatter(x=x, y=y, mode="lines+markers", name=name, marker_color=current_color)
    fig.add_trace(trace, 1, 1)

    df = hist_for_plot["current"].sort_values("x")
    x = [str(x) for x in df.x]
    y = list(df["count"])
    trace = go.Bar(name="current", x=x, y=y, marker_color=current_color)
    fig.add_trace(trace, 2, 1)

    if is_ref_data:
        sorted_index = val_for_plot["reference"].sort_index()
        x = [str(idx) for idx in sorted_index.index]
        y = list(sorted_index)
        trace = go.Scatter(x=x, y=y, mode="lines+markers", name=name, marker_color=reference_color)
        fig.add_trace(trace, 1, 1)

        df = hist_for_plot["reference"].sort_values("x")
        x = [str(x) for x in df.x]
        y = list(df["count"])
        trace = go.Bar(name="reference", x=x, y=y, marker_color=reference_color)
        fig.add_trace(trace, 2, 1)

    fig.update_yaxes(title_text=name, row=1, col=1)
    fig.update_yaxes(title_text="count", row=2, col=1)
    title = f"current {name}: {np.round(curr_metric, 3)}"

    if is_ref_data:
        title += f", reference {name}: {np.round(ref_metric, 3)}"

    fig.update_layout(title=title)
    return fig


def plot_value_counts_tables(feature_name, values, curr_df, ref_df, id_prfx):
    additional_plots = []
    if values is not None:
        curr_df = curr_df[curr_df["count"] != 0]
        curr_vals_inside_lst = curr_df[curr_df.x.isin(values)].sort_values("count", ascending=False)
        if curr_vals_inside_lst.shape[0] > 0:
            additional_plots.append(
                DetailsInfo(
                    title="Values inside the list (top 10)",
                    info=table_data(column_names=["value", "count"], data=curr_vals_inside_lst[:10].values),
                )
            )
        curr_vals_outside_lst = curr_df[~curr_df.x.isin(values)].sort_values("count", ascending=False)
        if curr_vals_outside_lst.shape[0] > 0:
            additional_plots.append(
                DetailsInfo(
                    title="Values outside the list (top 10)",
                    info=table_data(column_names=["value", "count"], data=curr_vals_outside_lst[:10].values),
                )
            )
    elif ref_df is not None:
        curr_df = curr_df[curr_df["count"] != 0]
        ref_df = ref_df[ref_df["count"] != 0]

        if is_numeric_dtype(curr_df.x):
            new_values = np.setdiff1d(curr_df.x.values, ref_df.x.values)
            missed_values = np.setdiff1d(ref_df.x.values, curr_df.x.values)
        else:
            curr_df["x"] = curr_df["x"].astype(str)
            ref_df["x"] = ref_df["x"].astype(str)
            new_values = np.setdiff1d(curr_df.x.values, ref_df.x.values)
            missed_values = np.setdiff1d(ref_df.x.values, curr_df.x.values)
        new_values_data = curr_df[curr_df.x.isin(new_values)].sort_values("count", ascending=False)
        missed_values_data = ref_df[ref_df.x.isin(missed_values)].sort_values("count", ascending=False)
        additional_plots.append(
            DetailsInfo(
                title="New values (top 10)",
                info=table_data(column_names=["value", "count"], data=new_values_data[:10].values),
            )
        )
        additional_plots.append(
            DetailsInfo(
                title="Missing values (top 10)",
                info=table_data(column_names=["value", "count"], data=missed_values_data[:10].values),
            )
        )

    return additional_plots


def plot_value_counts_tables_ref_curr(feature_name, curr_df, ref_df, id_prfx):
    additional_plots = []
    curr_df = curr_df[curr_df["count"] != 0]

    additional_plots.append(
        DetailsInfo(
            title="Current value counts (top 10)",
            info=table_data(column_names=["value", "count"], data=curr_df[:10].values),
        )
    )
    if ref_df is not None:
        ref_df = ref_df[ref_df["count"] != 0]
        additional_plots.append(
            DetailsInfo(
                title="Reference value counts (top 10)",
                info=table_data(column_names=["value", "count"], data=ref_df[:10].values),
            )
        )
    return additional_plots


def approx(value, relative=None, absolute=None):
    """Get approximate value for checking a value is equal to other within some tolerance"""
    return ApproxValue(value=value, relative=relative, absolute=absolute)


def dataframes_to_table(
    current: pd.DataFrame,
    reference: Optional[pd.DataFrame],
    columns: List[str],
    table_id: str,
    sort_by: str = "curr",
    na_position: str = "first",
    asc: bool = False,
):
    display_columns = ["display"]
    if reference is not None:
        display_columns += ["ref_display"]
        df = pd.merge(
            current,
            reference.rename(columns={"value": "ref_value", "display": "ref_display"}),
            how="outer",
            left_index=True,
            right_index=True,
        )
        df["eq"] = (df["value"] == df["ref_value"]) | (df["value"].isna() & df["ref_value"].isna())
        if "ref_display" not in df.columns:
            df["ref_display"] = df["ref_value"].fillna("NA").astype(str)
    else:
        df = current
    if "display" not in df.columns:
        df["display"] = df["value"].fillna("NA").astype(str)
    if sort_by == "curr":
        df.sort_values("value", na_position=na_position, inplace=True, ascending=asc)
    elif sort_by == "diff" and reference is not None:
        df.sort_values("eq", inplace=True)

    df = df[display_columns].fillna("NA")
    return [
        DetailsInfo(
            id=table_id,
            title="",
            info=BaseWidgetInfo(
                title="",
                type="table",
                params={"header": list(columns), "data": [[idx] + list(df.loc[idx].values) for idx in df.index]},
                size=2,
            ),
        )
    ]


def plot_dicts_to_table(
    dict_curr: dict,
    dict_ref: Optional[dict],
    columns: list,
    id_prfx: str,
    sort_by: str = "curr",
    asc: bool = False,
):
    return dataframes_to_table(
        pd.DataFrame.from_dict(dict_curr, orient="index", columns=["value"]),
        None if dict_ref is None else pd.DataFrame.from_dict(dict_ref, orient="index", columns=["value"]),
        columns=columns,
        table_id=id_prfx,
        sort_by=sort_by,
        na_position="first",
        asc=asc,
    )


def plot_correlations(current_correlations, reference_correlations):
    columns = current_correlations.columns
    heatmap_text = None
    heatmap_texttemplate = None

    if reference_correlations is not None:
        cols = 2
        subplot_titles = ["current", "reference"]
    else:
        cols = 1
        subplot_titles = [""]

    fig = make_subplots(rows=1, cols=cols, subplot_titles=subplot_titles, shared_yaxes=True)
    if len(columns) < 15:
        heatmap_text = np.round(current_correlations, 2).astype(str)
        heatmap_texttemplate = "%{text}"

    trace = go.Heatmap(
        z=current_correlations,
        x=columns,
        y=columns,
        text=heatmap_text,
        texttemplate=heatmap_texttemplate,
        coloraxis="coloraxis",
    )
    fig.add_trace(trace, 1, 1)

    if reference_correlations is not None:
        if len(columns) < 15:
            heatmap_text = np.round(reference_correlations, 2).astype(str)
            heatmap_texttemplate = "%{text}"

        trace = go.Heatmap(
            z=reference_correlations,
            x=columns,
            y=columns,
            text=heatmap_text,
            texttemplate=heatmap_texttemplate,
            coloraxis="coloraxis",
        )
        fig.add_trace(trace, 1, 2)
    fig.update_layout(coloraxis={"colorscale": "RdBu_r"})
    fig.update_yaxes(type="category")
    fig.update_xaxes(tickangle=-45)
    return fig


def plot_conf_mtrx(curr_mtrx, ref_mtrx):
    if ref_mtrx is not None:
        cols = 2
        subplot_titles = ["current", "reference"]
    else:
        cols = 1
        subplot_titles = [""]
    fig = make_subplots(rows=1, cols=cols, subplot_titles=subplot_titles, shared_yaxes=True)
    trace = go.Heatmap(
        z=curr_mtrx.values,
        x=curr_mtrx.labels,
        y=curr_mtrx.labels,
        text=np.array(curr_mtrx.values).astype(str),
        texttemplate="%{text}",
        coloraxis="coloraxis",
    )
    fig.add_trace(trace, 1, 1)

    if ref_mtrx is not None:
        trace = go.Heatmap(
            z=ref_mtrx.values,
            x=ref_mtrx.labels,
            y=ref_mtrx.labels,
            text=np.array(ref_mtrx.values).astype(str),
            texttemplate="%{text}",
            coloraxis="coloraxis",
        )
        fig.add_trace(trace, 1, 2)
    fig.update_layout(coloraxis={"colorscale": "RdBu_r"})
    return fig


def plot_roc_auc(
    *, curr_roc_curve: dict, ref_roc_curve: Optional[dict], color_options: ColorOptions
) -> List[Tuple[str, BaseWidgetInfo]]:
    additional_plots = []
    cols = 1
    subplot_titles = [""]
    current_color = color_options.get_current_data_color()
    reference_color = color_options.get_reference_data_color()

    if ref_roc_curve is not None:
        cols = 2
        subplot_titles = ["current", "reference"]

    for label in curr_roc_curve.keys():
        fig = make_subplots(rows=1, cols=cols, subplot_titles=subplot_titles, shared_yaxes=True)
        trace = go.Scatter(
            x=curr_roc_curve[label]["fpr"],
            y=curr_roc_curve[label]["tpr"],
            mode="lines",
            name="ROC",
            marker=dict(
                size=6,
                color=current_color,
            ),
        )
        fig.add_trace(trace, 1, 1)

        if ref_roc_curve is not None:
            trace = go.Scatter(
                x=ref_roc_curve[label]["fpr"],
                y=ref_roc_curve[label]["tpr"],
                mode="lines",
                name="ROC",
                marker=dict(
                    size=6,
                    color=reference_color,
                ),
            )
            fig.add_trace(trace, 1, 2)
        fig.update_layout(yaxis_title="True Positive Rate", xaxis_title="False Positive Rate", showlegend=True)
        additional_plots.append((f"ROC Curve for label {label}", plotly_figure(title="", figure=fig)))

    return additional_plots


def plot_boxes(*, curr_for_plots: dict, ref_for_plots: Optional[dict], color_options: ColorOptions):
    current_color = color_options.get_current_data_color()
    reference_color = color_options.get_reference_data_color()
    fig = go.Figure()
    trace = go.Box(
        lowerfence=curr_for_plots["mins"],
        q1=curr_for_plots["lowers"],
        q3=curr_for_plots["uppers"],
        median=curr_for_plots["means"],
        upperfence=curr_for_plots["maxs"],
        name="current",
        marker_color=current_color,
    )
    fig.add_trace(trace)
    if ref_for_plots is not None:
        trace = go.Box(
            lowerfence=curr_for_plots["mins"],
            q1=ref_for_plots["lowers"],
            q3=ref_for_plots["uppers"],
            median=ref_for_plots["means"],
            upperfence=ref_for_plots["maxs"],
            name="reference",
            marker_color=reference_color,
        )
        fig.add_trace(trace)
        fig.update_layout(boxmode="group")

    fig.update_layout(
        yaxis_title="Prerdictions",
        xaxis_title="Class",
    )
    return fig


def plot_rates(*, curr_rate_plots_data: dict, ref_rate_plots_data: Optional[dict] = None, color_options: ColorOptions):
    if ref_rate_plots_data is not None:
        cols = 2
        subplot_titles = ["current", "reference"]
    else:
        cols = 1
        subplot_titles = [""]

    curr_df = pd.DataFrame(
        {
            "thrs": curr_rate_plots_data["thrs"],
            "fpr": curr_rate_plots_data["fpr"],
            "tpr": curr_rate_plots_data["tpr"],
            "fnr": curr_rate_plots_data["fnr"],
            "tnr": curr_rate_plots_data["tnr"],
        }
    )
    curr_df = curr_df[curr_df.thrs <= 1]

    fig = make_subplots(rows=1, cols=cols, subplot_titles=subplot_titles, shared_yaxes=True)
    for i, metric in enumerate(["fpr", "tpr", "fnr", "tnr"]):
        fig.add_trace(
            go.Scatter(
                x=curr_df["thrs"],
                y=curr_df[metric],
                mode="lines",
                legendgroup=metric,
                name=metric,
                marker_color=color_options.color_sequence[i],
            ),
            1,
            1,
        )
    if ref_rate_plots_data is not None:
        ref_df = pd.DataFrame(
            {
                "thrs": ref_rate_plots_data["thrs"],
                "fpr": ref_rate_plots_data["fpr"],
                "tpr": ref_rate_plots_data["tpr"],
                "fnr": ref_rate_plots_data["fnr"],
                "tnr": ref_rate_plots_data["tnr"],
            }
        )
        ref_df = ref_df[ref_df.thrs <= 1]
        for i, metric in enumerate(["fpr", "tpr", "fnr", "tnr"]):
            fig.add_trace(
                go.Scatter(
                    x=ref_df["thrs"],
                    y=ref_df[metric],
                    mode="lines",
                    legendgroup=metric,
                    name=metric,
                    showlegend=False,
                    marker_color=color_options.color_sequence[i],
                ),
                1,
                2,
            )
    fig.update_layout(
        xaxis_title="threshold",
    )

    return fig



from datetime import datetime
from typing import Any
from typing import Dict

import pandas as pd

from evidently.analyzers.data_quality_analyzer import DataQualityAnalyzer
from evidently.calculations.data_quality import DataQualityStats
from evidently.model_profile.sections.base_profile_section import ProfileSection


class DataQualityProfileSection(ProfileSection):
    def part_id(self) -> str:
        return "data_quality"

    def __init__(self):
        super().__init__()
        self.analyzers_types = [DataQualityAnalyzer]
        self._result = None

    def analyzers(self):
        return self.analyzers_types

    @staticmethod
    def _get_stats_as_dict(all_features: DataQualityStats) -> Dict[str, Dict[str, Any]]:
        result: Dict[str, Dict[str, Any]] = {}

        for feature_name, feature_stats in all_features.get_all_features().items():
            result[feature_name] = {}

            for stat_name, stat_value in feature_stats.as_dict().items():
                if stat_value is not None:
                    result[feature_name][stat_name] = stat_value

        return result

    @staticmethod
    def _get_corr_matrices_as_dict(
        correlations: Dict[str, pd.DataFrame]
    ) -> Dict[str, Any]:
        result: Dict[str, Dict[str, Any]] = {}

        for kind, corr_df in correlations.items():
            result[kind] = {}
            for feature in corr_df.columns:
                result[kind][feature] = {
                    k: v
                    for (k, v) in zip(corr_df[feature].index, corr_df[feature])
                    if k != feature
                }

        return result

    def calculate(
        self, reference_data, current_data, column_mapping, analyzers_results
    ):
        result = DataQualityAnalyzer.get_results(analyzers_results)
        result_json = result.columns.as_dict()
        result_json["metrics"] = {}
        result_json["correlations"] = {}

        if result.reference_features_stats:
            result_json["metrics"]["reference"] = self._get_stats_as_dict(
                result.reference_features_stats
            )

        if result.current_features_stats:
            result_json["metrics"]["current"] = self._get_stats_as_dict(
                result.current_features_stats
            )

        if result.reference_correlations:
            result_json["correlations"]["reference"] = self._get_corr_matrices_as_dict(
                result.reference_correlations
            )

        if result.current_correlations:
            result_json["correlations"]["current"] = self._get_corr_matrices_as_dict(
                result.current_correlations
            )

        self._result = {
            "name": self.part_id(),
            "datetime": str(datetime.now()),
            "data": result_json,
        }

    def get_results(self):
        return self._result



import dataclasses
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import pandas as pd

from evidently.calculations import stattests

StatTestFuncType = Callable[[pd.Series, pd.Series, str, float], Tuple[float, bool]]


@dataclasses.dataclass
class StatTestResult:
    drift_score: float
    drifted: bool
    actual_threshold: float


@dataclasses.dataclass
class StatTest:
    name: str
    display_name: str
    func: StatTestFuncType
    allowed_feature_types: List[str]
    default_threshold: float = 0.05

    def __call__(
        self, reference_data: pd.Series, current_data: pd.Series, feature_type: str, threshold: Optional[float]
    ) -> StatTestResult:
        actual_threshold = self.default_threshold if threshold is None else threshold
        p = self.func(reference_data, current_data, feature_type, actual_threshold)
        drift_score, drifted = p
        return StatTestResult(drift_score=drift_score, drifted=drifted, actual_threshold=actual_threshold)

    def __hash__(self):
        # hash by name, so stattests with same name would be the same.
        return hash(self.name)


PossibleStatTestType = Union[str, StatTestFuncType, StatTest]

_registered_stat_tests: Dict[str, Dict[str, StatTest]] = {}
_registered_stat_test_funcs: Dict[StatTestFuncType, str] = {}


def register_stattest(stat_test: StatTest):
    _registered_stat_tests[stat_test.name] = {ft: stat_test for ft in stat_test.allowed_feature_types}
    _registered_stat_test_funcs[stat_test.func] = stat_test.name


def _get_default_stattest(reference_data: pd.Series, current_data: pd.Series, feature_type: str) -> StatTest:
    n_values = pd.concat([reference_data, current_data]).nunique()
    if feature_type == "text":
        return stattests.text_content_drift_stat_test
    elif reference_data.shape[0] <= 1000:
        if feature_type == "num":
            if n_values <= 5:
                return stattests.chi_stat_test if n_values > 2 else stattests.z_stat_test
            elif n_values > 5:
                return stattests.ks_stat_test
        elif feature_type == "cat":
            return stattests.chi_stat_test if n_values > 2 else stattests.z_stat_test
    elif reference_data.shape[0] > 1000:
        if feature_type == "num":
            n_values = pd.concat([reference_data, current_data]).nunique()
            if n_values <= 5:
                return stattests.jensenshannon_stat_test
            elif n_values > 5:
                return stattests.wasserstein_stat_test
        elif feature_type == "cat":
            return stattests.jensenshannon_stat_test
    raise ValueError(f"Unexpected feature_type {feature_type}")


def get_stattest(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, stattest_func: Optional[PossibleStatTestType]
) -> StatTest:
    if stattest_func is None:
        return _get_default_stattest(reference_data, current_data, feature_type)
    if isinstance(stattest_func, StatTest):
        return stattest_func
    if callable(stattest_func) and stattest_func not in _registered_stat_test_funcs:
        return StatTest(
            name="",
            display_name=f"custom function '{stattest_func.__name__}'",
            func=stattest_func,
            allowed_feature_types=[],
        )
    if callable(stattest_func) and stattest_func in _registered_stat_test_funcs:
        stattest_name = _registered_stat_test_funcs[stattest_func]
    elif isinstance(stattest_func, str):
        stattest_name = stattest_func
    else:
        raise ValueError(f"Unexpected type of stattest argument ({type(stattest_func)}), expected: str or Callable")
    funcs = _registered_stat_tests.get(stattest_name, None)
    if funcs is None:
        raise StatTestNotFoundError(stattest_name)
    func = funcs.get(feature_type)
    if func is None:
        raise StatTestInvalidFeatureTypeError(stattest_name, feature_type)
    return func


class StatTestNotFoundError(ValueError):
    def __init__(self, stattest_name: str):
        super().__init__(
            f"No stattest found of name {stattest_name}. " f"Available stattests: {list(_registered_stat_tests.keys())}"
        )


class StatTestInvalidFeatureTypeError(ValueError):
    def __init__(self, stattest_name: str, feature_type: str):
        super().__init__(
            f"Stattest {stattest_name} isn't applicable to feature of type {feature_type}. "
            f"Available feature types: {list(_registered_stat_tests[stattest_name].keys())}"
        )
